# SPDX-License-Identifier: MPL-2.0
# Copyright 2024, ASAM e.V.
# This Source Code Form is subject to the terms of the Mozilla
# Public License, v. 2.0. If a copy of the MPL was not distributed
# with this file, You can obtain one at https://mozilla.org/MPL/2.0/.
from __future__ import annotations
from typing import TYPE_CHECKING, Dict, List, Any

import json
from ..base import ReportFormatter
from qc_baselib import IssueSeverity
import xml.etree.ElementTree as ETree

if TYPE_CHECKING:
    from ..base.report_formatter import TStream
    from qc_baselib import Result
    from qc_baselib.models.common import ParamType
    from qc_baselib.models.result import CheckerType, MetadataType, RuleType, IssueType


class JsonFormatter(ReportFormatter):
    @property
    def module_name(self):
        return "qc-report-json"
    
    @property
    def default_custom_parameters(self):
        return {"Indent": 2}
    
    @property
    def default_output_file_param_value(self) -> str:
        return "Report.json"

    def _dump(self, result: "Result", output: "TStream"):
        indent = int(self.get_param("Indent"))  # type: ignore
        self.result = result

        self._addressed_rule_uids = set()
        self._violated_rule_uids = {
            IssueSeverity.ERROR: set(),
            IssueSeverity.WARNING: set(),
            IssueSeverity.INFORMATION: set(),
        }

        dump_ = {
            "version": self.result.get_result_version(),
            "bundles": self._dump_checker_bundle_results(),
            "rules": self._dump_addressed_and_violated_rules()
        }

        json.dump(dump_, output, indent=indent)
        
    def _dump_addressed_and_violated_rules(self) -> Dict[str, Any]:
        return {
            "addressed": list(self._addressed_rule_uids),
            "violated": dict(((key.name.lower(), list(value)) for key, value in self._violated_rule_uids.items()))
        }

    def _dump_checker_bundle_results(self) -> Dict[str, Any]:
        result = {}
        for bundle in self.result.get_checker_bundle_results():
            result[bundle.name] = {
                "name": bundle.name,
                "version": bundle.version,
                "summary": bundle.summary,
                "description": bundle.description,
                "parameters": self._dump_params(bundle.params),
                "checkers": self._dump_checkers(bundle.checkers),
            }
        return result

    def _dump_params(self, params: List["ParamType"]) -> Dict[str, Dict[str, Any]]:
        return dict(((p.name, {"name": p.name, "value": p.value}) for p in params))  # type: ignore

    def _dump_checkers(self, checkers: List["CheckerType"]) -> Dict[str, Any]:
        result = {}
        for checker in checkers:
            result[checker.checker_id] = {
                "checkerId": checker.checker_id,
                "description": checker.description,
                "summary": checker.summary,
                "status": str(checker.status.value),
                "metadata": self._dump_metadata(checker.metadata),
                "addressedRule": self._dump_addressed_rule(checker.addressed_rule),
                "issues": self._dump_issues(checker.issues),
            }
        return result

    def _dump_metadata(
        self, metadata: List["MetadataType"]
    ) -> Dict[str, Dict[str, Any]]:
        return dict(
            (m.key, {"key": m.key, "value": m.value, "description": m.description})
            for m in metadata
        )

    def _dump_addressed_rule(self, addressed_rule: List["RuleType"]) -> List[str]:
        result = set((r.rule_uid for r in addressed_rule))
        self._addressed_rule_uids = self._addressed_rule_uids.union(result)
        return list(result)

    def _dump_issues(self, issues: List["IssueType"]) -> List[Dict[str, Any]]:
        result = []
        for issue in issues:
            self._violated_rule_uids[issue.level].add(issue.rule_uid)
            result.append(
                {
                    "issueId": issue.issue_id,
                    "level": issue.level.name,
                    "description": issue.description,
                    "ruleUid": issue.rule_uid,
                    "locations": self._dump_locations(issue.locations),
                    "domainSpecificInfo": self._dump_domain_specific_info(
                        issue.domain_specific_info
                    ),
                }
            )
        return result

    def _dump_locations(self, locations):
        file_locations = []
        xml_locations = []
        inertial_locations = []
        time_locations = []
        message_locations = []

        for location in locations:
            for file_location in location.file_location:
                file_locations.append(
                    {
                        "description": location.description,
                        "column": file_location.column,
                        "row": file_location.row,
                        "offset": file_location.offset,
                    }
                )
            for xml_location in location.xml_location:
                xml_locations.append(
                    {
                        "description": location.description,
                        "xpath": xml_location.xpath,
                    }
                )
            for inertial_location in location.inertial_location:
                inertial_locations.append(
                    {
                        "description": location.description,
                        "x": inertial_location.x,
                        "y": inertial_location.y,
                        "z": inertial_location.z,
                    }
                )
            for time_location in location.time_location:
                time_locations.append(
                    {
                        "description": location.description,
                        "time": time_location.time,
                    }
                )
            for message_location in location.message_location:
                message_locations.append(
                    {
                        "description": location.description,
                        "index": message_location.index,
                        "channel": message_location.channel,
                        "field": message_location.field,
                        "time": message_location.time,
                    }
                )
        return {
            "file": file_locations,
            "xml": xml_locations,
            "inertial": inertial_locations,
            "time": time_locations,
            "message": message_locations,
        }

    def _dump_domain_specific_info(self, domain_specific_info):
        return [ETree.tostring(dsi, encoding="unicode") for dsi in domain_specific_info]



# SPDX-License-Identifier: MPL-2.0
# Copyright 2024, ASAM e.V.
# This Source Code Form is subject to the terms of the Mozilla
# Public License, v. 2.0. If a copy of the MPL was not distributed
# with this file, You can obtain one at https://mozilla.org/MPL/2.0/.
from __future__ import annotations
from typing import TYPE_CHECKING

import json
from ..base import ReportFormatter
from qc_baselib import IssueSeverity

if TYPE_CHECKING:
    from ..base.report_formatter import TStream
    from qc_baselib import Result

class GithubCIFormatter(ReportFormatter):
    def __init__(self, *args, **kwargs):
        self._has_issue = False
        super().__init__(*args, **kwargs)

    @property
    def module_name(self):
        return "qc-report-github-ci"
    
    @property
    def default_output_file_param_value(self) -> str:
        return "Report.txt"
    
    @property
    def has_issue(self) -> bool:
        return self._has_issue

    def _dump(self, result: "Result", output: "TStream"):
        self._has_issue = False

        for bundle in result.get_checker_bundle_results():
            for checker in bundle.checkers:
                for issue in checker.issues:
                    issue_level = issue.level.name.lower()
                    bundle_name = bundle.name
                    checker_id = checker.checker_id
                    rule_uid = issue.rule_uid
                    description = issue.description

                    report_line = f"{issue_level}: {bundle_name} / {checker_id} / {rule_uid}: {description}\n"
                    output.write(report_line)

                    if issue.level == IssueSeverity.ERROR:
                        self._has_issue = True



# SPDX-License-Identifier: MPL-2.0
# Copyright 2024, ASAM e.V.
# This Source Code Form is subject to the terms of the Mozilla
# Public License, v. 2.0. If a copy of the MPL was not distributed
# with this file, You can obtain one at https://mozilla.org/MPL/2.0/.
from .report_formatter import ReportFormatter
from .report_application import ReportApplication


# SPDX-License-Identifier: MPL-2.0
# Copyright 2024, ASAM e.V.
# This Source Code Form is subject to the terms of the Mozilla
# Public License, v. 2.0. If a copy of the MPL was not distributed
# with this file, You can obtain one at https://mozilla.org/MPL/2.0/.


def main():
    from ..base import ReportApplication
    from .text_formatter import TextFormatter

    formatter = TextFormatter()
    app = ReportApplication(
        formatter,
        prog="qc_report_text",
        description="Creates report in human readable text format")
    return app.main()


if __name__ == "__main__":
    main()


# SPDX-License-Identifier: MPL-2.0
# Copyright 2024, ASAM e.V.
# This Source Code Form is subject to the terms of the Mozilla
# Public License, v. 2.0. If a copy of the MPL was not distributed
# with this file, You can obtain one at https://mozilla.org/MPL/2.0/.
import os
import enum

from typing import List
from pydantic import BaseModel, field_validator


class ExecType(str, enum.Enum):
    EXECUTABLE = "executable"


class ModuleType(str, enum.Enum):
    CHECKER_BUNDLE = "checker_bundle"
    REPORT_MODULE = "report_module"
    RESULT_POOLING = "result_pooling"


class FrameworkManifest(BaseModel):
    manifest_file_path: List[str]

    @field_validator("manifest_file_path")
    @classmethod
    def file_path_must_exist(cls, v: List[str]) -> List[str]:
        for raw_file_path in v:
            resolved_file_path = os.path.expandvars(raw_file_path)
            if not os.path.isfile(resolved_file_path):
                raise ValueError(f"File path '{resolved_file_path}' must exist.")
        return v


class Module(BaseModel):
    name: str
    exec_type: ExecType
    module_type: ModuleType
    exec_command: str


class ModuleManifest(BaseModel):
    module: List[Module]



# SPDX-License-Identifier: MPL-2.0
# Copyright 2024, ASAM e.V.
# This Source Code Form is subject to the terms of the Mozilla
# Public License, v. 2.0. If a copy of the MPL was not distributed
# with this file, You can obtain one at https://mozilla.org/MPL/2.0/.
from __future__ import annotations

from argparse import ArgumentParser
import datetime
import os
import subprocess

from typing import TYPE_CHECKING, Dict

from qc_baselib import Configuration
from .models import FrameworkManifest, ModuleManifest, ModuleType

if TYPE_CHECKING:
    from .models import Module


FRAMEWORK_WORKING_DIR_VAR_NAME = "ASAM_QC_FRAMEWORK_WORKING_DIR"
FRAMEWORK_CONFIG_PATH_VAR_NAME = "ASAM_QC_FRAMEWORK_CONFIG_FILE"


def run_module_command(
    module: "Module", config_file_path: str, working_dir: str
) -> None:
    r"""
    Execute command specified in module configured with information in the
    provided configuration.

    Args:
        module: module to be executed
        config_file_path: defines the configuration file path
        working_dir: defines the working directory
    """
    try:
        print(f"Executing command: {os.path.expandvars(module.exec_command)}")

        cmd_env = os.environ.copy()
        cmd_env[FRAMEWORK_WORKING_DIR_VAR_NAME] = working_dir
        cmd_env[FRAMEWORK_CONFIG_PATH_VAR_NAME] = config_file_path

        process = subprocess.run(
            module.exec_command,
            env=cmd_env,
            shell=True,
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
        )

        exit_code = process.returncode
        if exit_code == 0:
            print("Command executed successfully.")
            print("Output:")
        else:
            print("Error occurred while executing the command.")
            print("Error message:")

        print(process.stdout.decode())

    except Exception as e:
        print(f"Error: {str(e)}")


def execute_modules(
    config_file_path: str,
    available_checker_bundles: Dict[str, "Module"],
    available_result_pooling: "Module",
    available_report_modules: Dict[str, "Module"],
    working_dir: str,
) -> None:
    r"""
    Execute the full sequence of operations, checker bundles, result
    pooling and result reporters as required as defined by the configuration file.
    Each module needs to be declared via manifest file.

    Args:
        config_file_path: path of the current configuration file
        available_checker_bundles: available bundle modules as declared in manifests
        available_result_pooling: result pooler module
        available_report_modules: available report modules as declared in manifests
        working_dir: working directory to be used for current execution
    """
    config = Configuration()
    config.load_from_file(config_file_path)

    checker_bundles = []
    report_modules = []

    for bundle in config.get_all_checker_bundles():
        name = bundle.application
        if name not in available_checker_bundles:
            raise RuntimeError(
                f"Checker bundle {name} is not available in the framework manifest."
            )

        checker_bundles.append(available_checker_bundles[name])

    for report in config.get_all_report_modules():
        name = report.application
        if name not in available_report_modules:
            raise RuntimeError(
                f"Report module {name} is not available in the framework manifest."
            )

        report_modules.append(available_report_modules[name])

    # Checker bundles
    print(f"Executing checker bundles")
    for checker_bundle in checker_bundles:
        print(f"Executing checker module:  {checker_bundle.name}")
        run_module_command(checker_bundle, config_file_path, working_dir)

    # Result pooling
    print(f"Executing result pooling:  {available_result_pooling.name}")
    run_module_command(available_result_pooling, config_file_path, working_dir)

    # Report modules
    print(f"Executing report modules")
    for report_module in report_modules:
        print(f"Executing report module:  {report_module.name}")
        run_module_command(report_module, config_file_path, working_dir)


def execute_runtime(
    config_file_path: str, manifest_file_path: str, working_dir: str
) -> None:
    r"""
    Execute all runtime operations defined in the input manifest over the
    defined configuration.

    Args:
        config_file_path: input configuration xml file path
        manifest_file_path: input manifest json file path
        working_dir: working directory
    """

    checker_bundles = {}
    report_modules = {}
    result_pooling = None

    with open(manifest_file_path, "rb") as framework_manifest_file:
        json_data = framework_manifest_file.read().decode()
        framework_manifest = FrameworkManifest.model_validate_json(json_data)

        for raw_module_manifest_path in framework_manifest.manifest_file_path:
            resolved_module_manifest_path = os.path.expandvars(raw_module_manifest_path)
            print(f"Registering manifest file at {resolved_module_manifest_path}")

            with open(resolved_module_manifest_path, "rb") as module_manifest_file:
                json_data = module_manifest_file.read().decode()
                module_manifest = ModuleManifest.model_validate_json(json_data)

                for module in module_manifest.module:
                    if module.module_type == ModuleType.CHECKER_BUNDLE:
                        if module.name in checker_bundles:
                            raise RuntimeError(
                                f"{module.name} is used by multiple checker bundles on framework manifest."
                            )
                        checker_bundles[module.name] = module
                    elif module.module_type == ModuleType.REPORT_MODULE:
                        if module.name in report_modules:
                            raise RuntimeError(
                                f"{module.name} is used by multiple report modules on framework manifest."
                            )
                        report_modules[module.name] = module
                    elif module.module_type == ModuleType.RESULT_POOLING:
                        if result_pooling is not None:
                            raise RuntimeError(
                                "Multiple result pooling modules defined on framework manifest. There must be exactly one result pooling module defined."
                            )
                        result_pooling = module
                    else:
                        raise RuntimeError(
                            f"Module of type {module.module_type.name} not supported."
                        )

    if result_pooling is None:
        raise RuntimeError(
            f"No result pooling module found in the provided framework manifest. There must be exactly one result pooling module defined."
        )

    execute_modules(
        config_file_path,
        checker_bundles,
        result_pooling,
        report_modules,
        working_dir,
    )


def main() -> None:
    r"""
    Runtime command line application
    """
    parser = ArgumentParser(
        prog="qc_runtime",
        description="Executes checker bundles for quality checking via configuration file",
        epilog="Quality Framework, Copyright 2024, ASAM e.V.: License MPLv2",
    )
    parser.add_argument(
        "--config",
        type=str,
        help="Path to the XML configuration file",
        required=True,
    )
    parser.add_argument(
        "--manifest",
        type=str,
        help="Path to the JSON manifest file",
        required=True,
    )

    parser.add_argument(
        "--working_dir",
        type=str,
        help="Working directory where all the output files are generated.",
        required=False,
    )

    args = parser.parse_args()

    working_dir = None
    if args.working_dir is None:
        formatted_now = datetime.datetime.now().strftime("%Y_%m_%d_%H_%M_%S_%f")
        working_dir = os.path.join(os.getcwd(), f"qc-output-{formatted_now}")
    else:
        working_dir = os.path.abspath(args.working_dir)

    os.makedirs(working_dir, exist_ok=True)

    execute_runtime(args.config, args.manifest, working_dir)


if __name__ == "__main__":
    main()



# SPDX-License-Identifier: MPL-2.0
# Copyright 2024, ASAM e.V.
# This Source Code Form is subject to the terms of the Mozilla
# Public License, v. 2.0. If a copy of the MPL was not distributed
# with this file, You can obtain one at https://mozilla.org/MPL/2.0/.
from argparse import ArgumentParser
from .report_formatter import ReportFormatter
import logging
import sys


class ReportApplication:
    r"""
    Class that wraps the common behavior for a ``ReportFormatter`` in order to create
    a command line application which is compatible with the design specification of the
    Quality Checker runtime.

    Application created with this utility class have the following command line
    arguments:

     * For each formatter parameters, a custom automatic optional argument is created
       programmatically, with declared default values.
     * create a specific configuration file for the current formatter. The configuration
       file will contain all the parameters supported. Default value for the parameter
       can be changed with command line input. It is possible to define the output file
     * report directly from a result file. It is possible to use the additional arguments
       to define the path of the input file
     * report from a configuration file. In this case, the precedence for the parameters
       follows:
        
        * lowest: default value
        * configuration file
        * highest: command line arguments

    It is possible to define stdout / stderr output by using the ``__stdout__`` / 
    ``__stderr__`` strings in command line arguments.

    There are optional arguments for the current class, provided as ``kwargs``.

    Use the ``main`` method to create the actual application.

    Example usage::

        def main():
            app = ReportApplication(formatter)
            return app.main()

        if __name__ == "__main__":
            main()

    :param formatter: formatter object for application creation
    :param prog: string that represent the name to be used in help
    :param description: description to include in the program help
    :param license: license epilogue to include in the program help
    """

    class Constants:
        DEFAULT_OUTPUT_CONFIG_FILE_NAME = f"Config.xml"
        INPUT_FILE_IS_STDIN = "__stdin__"
        OUTPUT_FILE_IS_STDERR = "__stderr__"
        OUTPUT_FILE_IS_STDOUT = "__stdout__"

    def __init__(self, formatter: ReportFormatter, *args, **kwargs):
        self._formatter = formatter

        _default_license = "Quality Framework, Copyright 2024, ASAM e.V.: License MPLv2"
        _default_description = f"Formatter {self._formatter.module_name} implementation"
        _default_prog = self._formatter.module_name

        self._log = logging.getLogger(self._formatter.module_name)

        self._args = ArgumentParser(
            prog=kwargs.get("prog") or _default_prog,
            description=kwargs.get("description") or _default_description,
            epilog=kwargs.get("license") or _default_license,
        )
        self._define_arguments()

    def main(self) -> int:
        r"""
        Main function. Executes the formatter.

        :return: integer value as exit number
        """
        try:
            args, override_params = self._parse_args_and_params()

            for name, value in override_params.items():
                self._formatter.set_param(name, value)

            if args.default_conf:
                self.export_default_config(args.default_conf_filename)
                return 0

            if args.from_result:
                self.report_from_result()
                return 0

            if args.from_config:
                self.report_from_config(args.from_config, override_params)
                return 0

        except Exception as e:
            self._log.exception(
                "Error while executing the report tool module: %s - %s",
                type(e).__name__,
                e,
            )
            return -1
        return 1

    def _define_arguments(self):
        r"""
        Defines application arguments. Parameters specific for the formatter
        are created as ``"param_{name of the parameter}"``
        """
        group = self._args.add_mutually_exclusive_group(required=True)
        group.add_argument(
            "-c",
            "--from-configuration",
            dest="from_config",
            type=str,
            help="run the application with an input configuration file",
        )
        group.add_argument(
            "-r",
            "--from-result",
            dest="from_result",
            action="store_true",
            help="run the application directly on an input result file",
        )
        group.add_argument(
            "-x",
            "--export-config",
            dest="default_conf",
            action="store_true",
            help="write default (or modified) configuration",
        )
        self._args.add_argument(
            "-xo",
            "--export-config-output",
            dest="default_conf_filename",
            default=self.Constants.DEFAULT_OUTPUT_CONFIG_FILE_NAME,
            help="optional name of file for default configuration",
        )

        # Automatically assigning parameters to command line argument
        class ArgsDefaulter:
            def __init__(self, default_type):
                self.default_type = default_type
            def __call__(self, arg):
                if arg is None:
                    return None
                return self.default_type(arg)
            
        param_group = self._args.add_argument_group("Parameters")
        for name in self._formatter.get_params_list():
            value = self._formatter.get_param(name)
            value_type = type(value)

            param_group.add_argument(
                f"--param-{name}",
                default=None,
                type=ArgsDefaulter(value_type),
                dest=f"param_{name}",
                help=f"formatter parameter {name} (default value: `{value}` [{value_type.__name__}])",
            )

    def _parse_args_and_params(self):
        r"""
        Evaluate command line parameters and extract the parameters override dictionary

        :return: a tuple of namespace and dictionary with override params
        """
        args = self._args.parse_args()

        override_params = {}
        for name in self._formatter.get_params_list():
            param_name = f"param_{name}"
            if param_name in args:
                value = getattr(args, param_name)
                if value:
                    override_params[name] = getattr(args, param_name)

        return args, override_params

    def export_default_config(self, output_filename):
        r"""
        Export the default configuration for the formatter. Command line
        parameter values can be used to customize the created configuration file.

        :param output_filename: name of the output file
        """
        from qc_baselib import Configuration
        from qc_baselib.models.config import Config, ReportModuleType
        from qc_baselib.models.common import ParamType

        parameters = [
            ParamType(name=name, value=self._formatter.get_param(name))
            for name in self._formatter.get_params_list()
        ]

        config = Config(
            params=[],
            checker_bundles=[],
            reports=[
                ReportModuleType(
                    application=self._formatter.module_name,
                    params=parameters,
                )
            ],
        )
        # FIXME: Fix this approach when public interface for reports is available
        config_string = config.to_xml(
            pretty_print=True, xml_declaration=True, standalone=False, encoding="utf-8"
        ).decode('utf-8')

        if output_filename == self.Constants.OUTPUT_FILE_IS_STDOUT:
            print(config_string, file=sys.stdout, flush=True)
            return
        if output_filename == self.Constants.OUTPUT_FILE_IS_STDERR:
            print(config_string, file=sys.stderr, flush=True)
            return
        with open(output_filename, "w") as output_fp:
            output_fp.write(config_string)  # type: ignore

    def report_from_result(self):
        r"""
        Runs the formatter by loading directly a result file
        """
        input: str = self._formatter.get_param(self._formatter.default_input_file_param_name)  # type: ignore
        output: str = self._formatter.get_param(self._formatter.default_output_file_param_name)  # type: ignore

        self._formatter.dump(input, output)

    def report_from_config(self, configuration_file, override_params):
        r"""
        Runs the formatter by loading a configuration file.

        .. info:: command line parameters override configuration parameter values

        :param configuration_file: path of the configuration file to load
        :param override_param: command line override params
        """
        from qc_baselib import Configuration
        from qc_baselib.models.config import Config

        configuration = Configuration()
        if configuration_file == self.Constants.INPUT_FILE_IS_STDIN:
            inner_config = Config.from_xml(sys.stdin.read().encode("utf-8"))
            # FIXME: Fix this for update public interface of models
            configuration._configuration = inner_config
        else:
            configuration.load_from_file(configuration_file)

        for name in self._formatter.get_params_list():
            if override_params.get(name) is not None:
                continue
            param_value = configuration.get_report_module_param(
                self._formatter.module_name, name
            )
            if param_value:
                self._formatter.set_param(name, param_value)

        self.report_from_result()



# SPDX-License-Identifier: MPL-2.0
# Copyright 2024, ASAM e.V.
# This Source Code Form is subject to the terms of the Mozilla
# Public License, v. 2.0. If a copy of the MPL was not distributed
# with this file, You can obtain one at https://mozilla.org/MPL/2.0/.
from __future__ import annotations
from typing import TYPE_CHECKING

from ..base import ReportFormatter
from qc_baselib import IssueSeverity
import xml.etree.ElementTree as ETree

if TYPE_CHECKING:
    from ..base.report_formatter import TStream, TParams
    from qc_baselib import Result


class TextFormatter(ReportFormatter):
    
    @property
    def module_name(self) -> str:
        return "qc-report-text"
    
    @property
    def default_output_file_param_value(self) -> str:
        return "Report.txt"
    
    @property
    def default_custom_parameters(self) -> "TParams":
        from os import get_terminal_size
        try:
            columns = get_terminal_size().columns
        except OSError:
            columns = 100
            
        return {
            "LineWidth": columns
        }

    def _dump(self, result: "Result", output: "TStream"):    
        columns = self.get_param("LineWidth")
        
        line_sep1 = ("=" * columns)
        line_sep2 = ("-" * columns)
        
        def p(val, indent=0):
            """
            Print with indent on the dump stream
            """
            if indent > 0:
                print(" " * indent, file=output, flush=True, end="")
            print(val, file=output, flush=True)
        
        def S(indent=0): 
            r"""
            Shortcut for chapter separator
            """
            p(line_sep1, indent)
        
        def s(indent=0):
            r"""
            Shortcut for section separator
            """
            p(line_sep2, indent)

        def n(indent=0):
            r"""
            Shortcut for empty line inclusion
            """
            p("", indent)
        
        addressed_rules = set()
        violated_rules = {
            IssueSeverity.ERROR: set(),
            IssueSeverity.WARNING: set(),
            IssueSeverity.INFORMATION: set(),
        }

        p(f"ASAM QUALITY FRAMEWORK v{result.get_result_version()} - RESULT REPORT")
        S()
        n()
        
        p("BUNDLES")
        s()
        n()
        
        bundles = result.get_checker_bundle_results()
        for bundle in bundles:
            p(f" -> {bundle.name} @ {bundle.version or 'version-undefined'}")
            p(f"Description: {bundle.description}", 4)
            p(f"Summary: {bundle.summary}", 4)
            if len(bundle.params):
                p(f"Parameters: ", 4)
                for param in bundle.params:
                    p(f" -> {param.name}: {param.value} [{type(param.value).__name__}]", 4)
            if len(bundle.checkers):
                p(f"Checkers: ", 4)
                for checker in bundle.checkers:
                    p(f" -> [{checker.status.name}] {checker.checker_id}", 4)
                    p(f"Description: {checker.description}", 8)
                    p(f"Summary: {checker.summary}", 8)
                    if len(checker.addressed_rule):
                        p("Addressed Rules:", 8)
                        for rule in checker.addressed_rule:
                            addressed_rules.add(rule.rule_uid)
                            p(f" -> {rule.rule_uid}", 8)
                    if len(checker.issues):
                        p("Issues:", 8)
                        for issue_no, issue in enumerate(checker.issues):
                            violated_rules[issue.level].add(issue.rule_uid)
                            p(f" {issue_no+1: 3d}. [{issue.level.name}] @ {issue.rule_uid}", 8)
                            p(f"Description: {issue.description}", 14)
                            if len(issue.domain_specific_info):
                                p("Domain Specific Info:", 14)
                                for dsi in issue.domain_specific_info:
                                    dsi_lines = ETree.tostring(dsi, encoding="unicode").split("\n")
                                    p(f" -> {dsi_lines[0].strip()}", 14)
                                    for dsi_line in dsi_lines[1:]:
                                        p(f"    {dsi_line.rstrip()}", 14)
                            if len(issue.locations):
                                p("Locations:", 14)
                                for location in issue.locations:
                                    for file_loc in location.file_location:
                                        p(f" -> File @ col: {file_loc.column or '-'}, row: {file_loc.row or '-'}, offset: {file_loc.offset or '-'}: {location.description}", 14)
                                    for xml_loc in location.xml_location:
                                        p(f" -> XML @ `{xml_loc.xpath}`: {location.description}", 14)
                                    for inertial_loc in location.inertial_location:
                                        p(f" -> Inertial @ [{inertial_loc.x:.4n}; {inertial_loc.y:.4n}; {inertial_loc.z:.4n}]: {location.description}", 14)
                                    for time_loc in location.time_location:
                                        p(f" -> Time @ {time_loc.time}: {location.description}", 14)
                                    for message_loc in location.message_location:
                                        p(f" -> Message @ index: {message_loc.index}, channel: {message_loc.channel or '-'}, field: {message_loc.field or '-'}, time: {message_loc.time or '-'}: {location.description}", 14)
                    if len(checker.metadata):
                        p("Metadata:", 8)
                        for metadata in checker.metadata:
                            p(f" -> {metadata.key}: {metadata.value} ({metadata.description})", 8)
                    n()
            if bundle.name != bundles[-1].name:
                s()
            n()
        S()
        n()
        p("ADDRESSED RULES")
        s()
        n()
        p(f"Addressed rules: {len(addressed_rules)}")
        for r in addressed_rules:
            p(f" -> {r}")
        n()

        p(f"Flagged Rules: {sum((len(x) for x in violated_rules.values()))}")
        for t, rs in violated_rules.items():
            p(f"with {t.name}: {len(rs)}", 2)
            for r in rs:
                p(f" -> {r}", 2)

        s()
        n()
        p("NOTES")
        s()
        n()
        p("Rule UID format:")
        p("  <emanating-entity>:<standard>:x.y.z:rule_set.for_rules.rule_name")
        n()
        p("where    ")
        p("  * Emanating Entity: a domain name for the entity (organization or company) that declares the rule UID")
        p("  * Standard: a short string that represents the standard or the domain to which the rule is applied")
        p("  * Definition Setting: the version of the standard or the domain to which the rule appears or is applied for the first time")
        p("  * Rule Full Name: the full name of the rule, as dot separated, snake lower case string. ")
        p("    The full name of a rule is composed by the rule set, a categorization for the rule, ")
        p("    and the rule name, a unique string inside the categorization. ")
        p("    The rule set can be nested (meaning that can be defined as an ")
        p("    arbitrary sequence of dot separated names, while the name is the snake ")
        p("    case string after the last dot of the full name)")


# SPDX-License-Identifier: MPL-2.0
# Copyright 2024, ASAM e.V.
# This Source Code Form is subject to the terms of the Mozilla
# Public License, v. 2.0. If a copy of the MPL was not distributed
# with this file, You can obtain one at https://mozilla.org/MPL/2.0/.
from __future__ import annotations
from typing import Optional, Union, Dict, List
from io import IOBase, StringIO, TextIOBase
from qc_baselib import Result
import sys


TParamValue = Optional[Union[str, int, float, bool]]
TParams = Dict[str, TParamValue]
TResultArg = Union[str, IOBase, Result]
TOutputArg = Optional[Union[str, IOBase, StringIO, TextIOBase]]
TStream = Union[IOBase, StringIO, TextIOBase]


class ReportFormatter:
    r"""
    Base interface declaration for a formatter class. It is possible to
    use this class in two way, that mainly deal in implementation details:

     * extension via inheritance (and override for existing methods). In 
       this case the implementers **must** override the function ``_dump``
       provided by the current class. Constructor must define required 
       properties for a correct extension
     * extension via dependency injection. In this case the implementers
       **must** provide a ``dump_function`` in constructor. The interface for the 
       function is the following::

           def dump_function(
               result: Result, 
               get_param: Callable[[str], Optional[TParamValue]], 
               output: TStream) -> None:
               pass

       where:

       :param result: the ``qc_baselib.Result`` object to convert
       :param get_param: a callable function that given a parameter name returns the value
       :param output: IO stream object where result should be written

    All reporter formatter are expected to have input and output parameters. Those parameters
    are already defined by the base class, with the following keys:

     * ``"InputFile"``: ``"Result.xqar"``
     * ``"OutputFile"``: ``"Report.txt"``

    the value can be modified by using injection or inheritance. To inject new values, simply
    provide new values via ``kwargs``. To change via inheritance, override the appropriate property.
    A full list of available ``kwargs`` is provided below.

    Additional parameters can be defined via injection (``kwargs``) or by an override of the property.
    The custom parameters can be used also to modify the default parameters.

    Example for a custom formatter via inheritance is provided below::

        class CustomFormatter(BaseFormatter):
            def __init__(self):
                super().__init__()

            def _dump(self, result: "Result", output: TStream):
                # Parameter example
                count_spaces = self.get_param("CountSpaces")  # Note: self.!
                output.write(" " * count_spaces)
                # Result dump example
                output.write(str(result))

            @property
            def module_name(self) -> str:
                return "custom-formatter"

            @property
            def default_output_file_param_value(self) -> str:
                return "CustomReport.bin"

            @property
            def default_custom_parameters(self) -> TParams:
                return {"CountSpaces": 2}

        formatter = CustomFormatter()
        formatter.dump("__stdin__", "__stdout__")

    Example for an identical injected formatter is provided below::

        def custom_dumper(result: "Result", get_param: Callable[[str], Optional[TParamValue]], output: TStream):
            # Parameter example
            count_spaces = get_param("CountSpaces")  # Note: no self.!
            output.write(" " * count_spaces)
            # Result dump example
            output.write(str(result))

        def new_custom_formatter():
            default_custom_parameters = {"CountSpaces": 2}
            return BaseFormatter(module_name="custom-formatter",
                dump_function=custom_dumper,
                default_output_file_param_value="CustomReport.bin",
                default_custom_parameters=default_custom_parameters)

        formatter = new_custom_formatter()
        formatter.set_param("CountSpaces", 4)
        formatter.dump("__stdin__", "__stdout__")

    The two implementation have virtually identical behavior.

    .. info:: It is possible also to use a mix of the two approaches

    From a user perspective, the only important methods are the constructor and the 
    ``dump`` method. ``dump`` method supports some placeholder string filename to integrate
    support for ``stdin`` as input, ``stdout`` and ``stderr`` as output, in place of 
    existing file.

    Parameters are exposed via the methods: ``get_params_list``, ``get_params``, ``set_params``.

    :param module_name: optional ``kwargs`` to define the module name. Should be a string 
    :param dump_function: optional ``kwargs`` to define the dump function. Should be a valid
                          callable with the aforementioned signature
    :param default_input_file_param_name: optional ``kwargs`` to change parameter input file key
                                         (default is ``"InputFile"``). Should be a string.
    :param default_input_file_param_name: optional ``kwargs`` to change parameter input file value
                                         (default is ``"Result.xqar"``). Should be a string.
    :param default_output_file_param_name: optional ``kwargs`` to change parameter output file key
                                         (default is ``"OutputFile"``). Should be a string.
    :param default_output_file_param_name: optional ``kwargs`` to change parameter output file value
                                         (default is ``"Report.txt"``). Should be a string.
    :param default_custom_parameters: optional ``kwargs`` for the definition of additional parameters.
                                      (default is ``{}``). With the same ``kwargs`` it is possible
                                      to override the default parameters value (but not keys)
    """

    class Constants:
        DEFAULT_INPUT_FILE_PARAM_NAME = "InputFile"
        DEFAULT_OUTPUT_FILE_PARAM_NAME = "OutputFile"
        DEFAULT_INPUT_FILE_PARAM_VALUE = "Result.xqar"
        DEFAULT_OUTPUT_FILE_PARAM_VALUE = f"Report.txt"
        INPUT_FILE_IS_STDIN = "__stdin__"
        OUTPUT_FILE_IS_STDERR = "__stderr__"
        OUTPUT_FILE_IS_STDOUT = "__stdout__"

    def __init__(self, *args, **kwargs):
        r"""
        Default constructor. Can be used for injection, or used in inherited constructors.

        Supported parameters defined in class documentation. 
        """
        self._module_name = kwargs.get("module_name") or type(self).__name__
        self._dump_function = kwargs.get("dump_function")

        for name in ["default_input_file_param_name", "default_input_file_param_value", 
                    "default_output_file_param_name", "default_output_file_param_value"]:
            setattr(self, f"_{name}", kwargs.get(name) or getattr(self.Constants, name.upper()))
        
        self._default_custom_parameters = (
            kwargs.get("default_custom_parameters") or {}
        )

        self._parameters = self.default_parameters()

    def set_param(self, name: str, value: TParamValue) -> "ReportFormatter":
        r"""
        Method to set parameters value for the current formatter. Only parameters
        already defined can be changed. It currently silently fails if name is 
        not available. This behavior may change in the future.

        .. warning:: There is no validation system in place for parameters!

        :param name: name of the parameter to set
        :param value: value to set for the parameter
        :return: this instance: the method is chainable
        """
        if name in self._parameters:
            self._parameters[name] = value
        return self
    
    def get_params_list(self) -> List[str]:
        r"""
        Returns a list of available parameters names

        :return: List of parameters names
        """
        return list(self._parameters.keys())

    def get_param(self, key: str) -> Optional[TParamValue]:
        r"""
        Get current value for a parameter. Silently fails if parameter name 
        is not available

        :param name: name of the parameter to query
        :return: value of the parameter or None
        """
        return self._parameters.get(key)

    def dump(
        self, input: TResultArg, output: TOutputArg = None
    ) -> Optional[str]:
        r"""
        Dumps the content of a result file in a different format. Input arguments
        controls both the loading and the data emission strategy.

        Regarding result loading:

         * if ``input`` is a ``str``, the result file is loaded from a file. If it 
           contains the specific placeholder ``__stdin__``, the content of the result
           file is consumed from the ``stdin`` stream.
         * if ``input`` is a ``io.IOBase``, content is consumed to be loaded inside the
           appropriate result object.
         * if ``input`` already contains a ``Result`` object, the object is used as is.

        Regarding the report output:

         * if ``output`` parameter is ``None``, the method returns a string
         * if ``output`` parameter is a ``str``, the result is written in a new
           file with ``output`` as filename. If the value contains the stdout or 
           stderr placeholder names (by default ``__stdout__`` and ``__stderr__``),
           those file are used for emitting the report.
         * if ``output`` is an ``io.IOBase`` compatible object (stream), the result is
           written directly to the stream.

        :param input: the result object to dump, or the filename as string, or a stream object
        :param output: optional output argument, as a string, or as a stream object
        :return: a string if ``output`` is None, None otherwise
        """
        result = self._result_loader(input)

        # Export to output
        if output is None:
            with StringIO() as string:
                self._dump(result, string)
                return string.getvalue()

        if isinstance(output, IOBase):
            self._dump(result, output)

        if isinstance(output, str):
            if output == self.Constants.OUTPUT_FILE_IS_STDOUT:
                self._dump(result, sys.stdout)  # type: ignore
            if output == self.Constants.OUTPUT_FILE_IS_STDERR:
                self._dump(result, sys.stderr)  # type: ignore
            with open(output, "w") as file_pointer:
                self._dump(result, file_pointer)

    def default_parameters(self) -> TParams:
        r"""
        Declares the default parameter object. Includes custom parameters.

        :return: dictionary with default parameters values
        """
        base_params: TParams = {
            self.default_input_file_param_name: self.default_input_file_param_value,
            self.default_output_file_param_name: self.default_output_file_param_value,
        }
        base_params.update(self.default_custom_parameters)
        return base_params
    
    def _result_loader(self, input: TResultArg) -> Result:
        r"""
        Load result file or stream
        """
        import io

        if isinstance(input, Result):
            return input
        
        result = Result()

        if isinstance(input, str):
            if input == self.Constants.INPUT_FILE_IS_STDIN:
                from qc_baselib.models.result import CheckerResults
                # FIXME: this should be fixed when public interface for Result is completed
                checker_results = CheckerResults.from_xml(sys.stdin.read().encode("utf-8"))
                result._report_results = checker_results
            else:
                result.load_from_file(input)
            return result
        

        if isinstance(input, io.IOBase):
            from qc_baselib.models.result import CheckerResults
            # FIXME: this should be fixed when public interface for Result is completed
            checker_results = CheckerResults.from_xml(input.read().encode("utf-8"))
            result._report_results = checker_results
        
        return result

    def _dump(self, result: "Result", output: TStream):
        r"""
        Dump logic. Input is a result object, output is the provided stream.

        .. info:: Override this function to create a custom report formatter with custom
                  dump logic from Result objet to text or binary data.

        .. warning:: This function shall not open or close the output stream

        :param result: result object to dump
        :param output: destination for the dump as an IO object (use ``write`` method)
        """
        if self._dump_function is None:
            raise NotImplementedError("Missing dump function for the current implementation")
        self._dump_function(result, self.get_param, output)

    @property
    def module_name(self) -> str:
        r"""
        Defines the module name

        .. info:: override this property to customize via inheritance the report formatter

        :return: module name to be used in default config
        """
        return self._module_name
    
    @property
    def default_input_file_param_name(self):
        r"""
        Defines the default input file name parameter name. Defaults to ``"InputFile"``

        .. info:: override this property to customize via inheritance the report formatter

        :return: input parameter name
        """
        return self._default_input_file_param_name  # type: ignore (dynamically defined)
    
    @property
    def default_input_file_param_value(self):
        r"""
        Defines the default input file name parameter value. Defaults to ``"Result.xqar"``

        .. info:: override this property to customize via inheritance the report formatter

        :return: input parameter value
        """
        return self._default_input_file_param_value  # type: ignore (dynamically defined)
    
    @property
    def default_output_file_param_name(self):
        r"""
        Defines the default output file name parameter name. Defaults to ``"OutputFile"``

        .. info:: override this property to customize via inheritance the report formatter

        :return: output parameter name
        """
        return self._default_output_file_param_name  # type: ignore (dynamically defined)
    
    @property
    def default_output_file_param_value(self):
        r"""
        Defines the default output file name parameter value. Defaults to ``"Report.txt"``

        .. info:: override this property to customize via inheritance the report formatter

        :return: output parameter value
        """
        return self._default_output_file_param_value  # type: ignore (dynamically defined)
    
    @property
    def default_custom_parameters(self) -> TParams:
        r"""
        Defines custom parameters default values for formatter

        .. info:: override this property to customize via inheritance the report formatter

        :return: dictionary of custom parameters
        """
        return self._default_custom_parameters  # type: ignore


# SPDX-License-Identifier: MPL-2.0
# Copyright 2024, ASAM e.V.
# This Source Code Form is subject to the terms of the Mozilla
# Public License, v. 2.0. If a copy of the MPL was not distributed
# with this file, You can obtain one at https://mozilla.org/MPL/2.0/.


def main():
    from ..base import ReportApplication
    from .github_ci_formatter import GithubCIFormatter

    class GithubCIApplication(ReportApplication):
        def __init__(self, formatter: GithubCIFormatter, *args, **kwargs):
            self.return_issues = False
            super().__init__(
                formatter,
                prog="qc_report_github_ci",
                description="Creates report to be used in GithubCI pipelines. Exit with error for issues",
                *args,
                **kwargs
            )

        def main(self):
            exit_value = super().main()
            if self.return_issues:
                return 2 if self._formatter.has_issue else 0  # type: ignore
            return exit_value

        def report_from_result(self):
            self.return_issues = True
            return super().report_from_result()

    formatter = GithubCIFormatter()
    app = GithubCIApplication(formatter)
    return app.main()


if __name__ == "__main__":
    main()
